{ #(@)$Id: GUITesting.pas,v 1.2 2000/12/05 01:15:40 juanco Exp $ }
{: DUnit: An XTreme testing framework for Delphi programs.
   @author  The DUnit Group.
   @version $Revision: 1.2 $
}
(*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is DUnit.
 *
 * The Initial Developers of the Original Code are Serge Beaumont
 * and Juancarlo Aez.
 * Portions created The Initial Developers are Copyright (C) 1999-2000.
 * Portions created by The DUnit Group are Copyright (C) 2000.
 * All rights reserved.
 *
 * Contributor(s):
 * Serge Beaumont <beaumose@iquip.nl>
 * Juanco Aez <juanco@users.sourceforge.net>
 * The DUnit group at SourceForge <http://dunit.sourceforge.net>
 *
 *)
unit GUITesting;

interface
uses
  TestFramework,

  SysUtils,
  Windows,
  Messages,
  Classes,
  Controls,
  Forms;

const
  rcs_id: string = '#(@)$Id: GUITesting.pas,v 1.2 2000/12/05 01:15:40 juanco Exp $';

type
  TGUITestCase = class(TTestCase)
  public
    constructor Create(MethodName :string); override;
  protected
    FGUI         :TControl; // this is the control we're testing
    FActionDelay :Integer;

    function  FindControl(Comp: TComponent; const CtlName: string): TControl; overload;
    function  FindControl(const Name: string): TControl;                      overload;

    function  FindParentWinControl(Control :TControl):TWinControl;

    function  ShiftStateToKeyData(ShiftState :TShiftState):Longint;
    procedure ClickLeftMouseButtonOn(Control: TControl);

    procedure Click;                      overload;
    procedure Click(ControlName :string); overload;
    procedure Click(control :TControl);   overload;

    procedure EnterKey(Key :Word; const ShiftState :TShiftState = []); overload;
    procedure EnterKeyInto(Control :TControl;   Key :Word; const ShiftState :TShiftState = []); overload;
    procedure EnterKeyInto(ControlName :string; Key :Word; const ShiftState :TShiftState = []); overload;

    procedure EnterKey(Key :Char; const ShiftState :TShiftState = []); overload;
    procedure EnterKeyInto(Control :TControl;   Key :Char; const ShiftState :TShiftState = []); overload;
    procedure EnterKeyInto(ControlName :string; Key :Char; const ShiftState :TShiftState = []); overload;

    procedure EnterText(Text :string);
    procedure EnterTextInto(Control :TControl;   Text :string); overload;
    procedure EnterTextInto(ControlName :string; Text :string); overload;

    procedure Show(OnOff :boolean = true);                      overload;
    procedure Show(Control :TControl; OnOff :boolean = true);   overload;
    procedure Show(ControlName :string; OnOff :boolean = true); overload;

    procedure Hide(OnOff :boolean = true);                      overload;
    procedure Hide(Control :TControl; OnOff :boolean = true);   overload;
    procedure Hide(ControlName :string; OnOff :boolean = true); overload;

    procedure Tab(n :Integer =1);


    procedure CheckTabTo(Control :TControl; Msg :string = '');  overload;
    procedure CheckTabTo(ControlName :string; Msg :string = '');  overload;

    function  GetFocused :TControl;
    procedure CheckFocused(Control :TControl);     overload;
    procedure CheckFocused(ControlName :string);   overload;

    property GUI         :TControl read FGUI write FGUI;
    property ActionDelay :Integer  read FActionDelay write FActionDelay;
  end;


implementation

// assertions are always on so we can check for own consistency
{$ASSERTIONS ON}

{ TGUITestCase }

constructor TGUITestCase.Create(MethodName :string);
begin
  inherited Create(MethodName);
  FActionDelay := 100;
end;

function TGUITestCase.FindControl(Comp: TComponent; const CtlName: string): TControl;

  function DoFind(C :TComponent; const CName :string) :TControl;
  var
    i: Integer;
  begin
    Result := nil;
    i := 0;
    while (Result = nil) and (i < C.ComponentCount) do
    begin
      with C do
      begin
        if (Components[i] is TControl)
        and (UpperCase(Components[i].Name) = CName) then
          Result := Components[I] as TControl
        else
          Result := DoFind(Components[I], CName);
      end;
      Inc(i);
    end;
  end;
begin
  Assert(Trim(CtlName) <> '', 'No control name');

  Result := DoFind(Comp, UpperCase(CtlName));

  Assert(Result <> nil, Format('Control named "%s" not found', [CtlName]));
end;

procedure TGUITestCase.ClickLeftMouseButtonOn(Control: TControl);
var
  P      :TSmallPoint;
begin
  Assert(Control <> nil, 'No control');

  Control := FindParentWinControl(Control);
  if Control <> nil then
  begin
    {:@ todo consider if this method should have X,Y parameters.
      @todo This doesn't work if the original control is not a TWinControl and is not in
       the middle of its parent. }
    P   := SmallPoint(Control.Width  div 2, Control.Height div 2);
    PostMessage(TWinControl(Control).Handle, WM_LBUTTONDOWN, 0, Longint(P));
    PostMessage(TWinControl(Control).Handle, WM_LBUTTONUP, 0,   Longint(P));
    Sleep(ActionDelay);
  end;
  Application.ProcessMessages;
end;

function TGUITestCase.FindControl(const Name: string): TControl;
begin
  Result := FindControl(GUI, Name);
end;

function TGUITestCase.FindParentWinControl(Control: TControl): TWinControl;
begin
  while (Control <> nil) and not (Control is TWinControl) do
    Control := Control.Parent;
  Result := TWinControl(Control);
end;

function TGUITestCase.ShiftStateToKeyData(ShiftState :TShiftState):Longint;
const
  AltMask = $20000000;
begin
  Result := 0;
  if ssShift in ShiftState then
    Result := Result or VK_SHIFT;
  if ssCtrl in ShiftState then
    Result := Result or VK_CONTROL;
  if ssAlt in ShiftState then
    Result := Result or AltMask;
end;

procedure TGUITestCase.Click;
begin
  Assert(GUI <> nil, 'GUI variable not set');
  Click(GUI);
end;

procedure TGUITestCase.Click(ControlName: string);
begin
  Click(FindControl(ControlName));
end;

procedure TGUITestCase.Click(Control: TControl);
begin
  Assert(Control <> nil, 'No control');
  ClickLeftMouseButtonOn(Control);
end;

procedure TGUITestCase.EnterKey(Key: Word; const ShiftState: TShiftState);
begin
  Assert(GUI <> nil, 'GUI variable not set');
  EnterKeyInto(GUI, Key, ShiftState);
end;

procedure TGUITestCase.EnterKeyInto(Control: TControl; Key: Word; const ShiftState: TShiftState);
begin
  Assert(Control <> nil, 'No control');
  Control := FindParentWinControl(Control);
  if Control <> nil then
  begin
    PostMessage(TWinControl(Control).Handle, WM_KEYDOWN, Key, ShiftStateToKeyData(ShiftState));
    Sleep(ActionDelay);
    PostMessage(TWinControl(Control).Handle, WM_KEYUP,   Key, ShiftStateToKeyData(ShiftState));
    Sleep(ActionDelay);
    Application.ProcessMessages;
  end;
end;

procedure TGUITestCase.EnterKeyInto(ControlName: string; Key: Word; const ShiftState: TShiftState);
begin
  EnterKeyInto(FindControl(ControlName), Key, ShiftState);
end;

procedure TGUITestCase.EnterKey(Key: Char; const ShiftState: TShiftState);
begin
  EnterKey(Ord(Key), ShiftState);
end;

procedure TGUITestCase.EnterKeyInto(Control: TControl; Key: Char; const ShiftState: TShiftState);
begin
  EnterKeyInto(Control, Ord(Key), ShiftState);
end;

procedure TGUITestCase.EnterKeyInto(ControlName: string; Key: Char; const ShiftState: TShiftState);
begin
  EnterKeyInto(ControlName, Ord(Key), ShiftState);
end;

procedure TGUITestCase.EnterText(Text: string);
begin
  Assert(GUI <> nil, 'GUI variable not set');
  EnterTextInto(GUI, Text);
end;

procedure TGUITestCase.EnterTextInto(ControlName, Text: string);
begin
  EnterTextInto(FindControl(ControlName), Text);
end;

procedure TGUITestCase.EnterTextInto(Control: TControl; Text: string);
var
  i :Integer;
begin
  Assert(Control <> nil, 'No control');
  Control := FindParentWinControl(Control);
  if Control <> nil then
  begin
    for i := 1 to Length(Text) do
    begin
      PostMessage(TWinControl(Control).Handle, WM_CHAR, Ord(Text[i]), 0);
      Sleep(ActionDelay);
    end;
  end;
  Application.ProcessMessages;
end;

procedure TGUITestCase.Show(OnOff: boolean);
begin
  Assert(GUI <> nil, 'GUI variable not set');
  Show(GUI, OnOff);
end;

procedure TGUITestCase.Show(ControlName: string; OnOff: boolean);
begin
  Show(FindControl(ControlName), OnOff);
end;

procedure TGUITestCase.Show(Control: TControl; OnOff: boolean);
begin
  Assert(Control <> nil);
  Control.Visible := OnOff;
  Assert(Control.Visible = OnOff);
  Application.ProcessMessages;
  Sleep(ActionDelay);
end;

procedure TGUITestCase.Hide(OnOff: boolean);
begin
  Show(false);
end;

procedure TGUITestCase.Hide(ControlName: string; OnOff: boolean);
begin
  Show(ControlName, false);
end;

procedure TGUITestCase.Hide(Control: TControl; OnOff: boolean);
begin
  Show(Control, false);
end;

procedure TGUITestCase.Tab(n: Integer);
var
  i :Integer;
  s :TShiftState;
begin
  Assert(GUI <> nil, 'GUI variable not set');

  s := [];
  if n < 0 then
  begin
    s := [ssShift];
    n := -n;
  end;

  for i := 1 to n do
    EnterKey(VK_TAB);

  Application.ProcessMessages;
  Sleep(ActionDelay);
end;

function TGUITestCase.GetFocused: TControl;
var
  i :Integer;
begin
  Assert(GUI <> nil, 'GUI variable not set');

  if GUI is TCustomForm then
    Result := TCustomForm(GUI).ActiveControl
  else
  begin
    Result := nil;
    for i := 0 to GUI.ComponentCount-1 do
    begin
       if (GUI.Components[i] is TWinControl)
       and TWinControl(GUI.Components[i]).Focused then
       begin
         Result := TControl(GUI.Components[i]);
         BREAK;
       end
    end;
  end;
end;



procedure TGUITestCase.CheckTabTo(Control: TControl; Msg :string = '');
var
  i :Integer;
begin
  Assert(GUI <> nil, 'GUI variable not set');

  Check(Control is TWinControl,
        Format('%s: Expected a TWinControl, but %s is a %s',
               [msg, Control.Name, Control.ClassName])
        );
  Check(TWinControl(Control).CanFocus,
        Format('%s: Control %s:%s cannot focus', [msg, Control.Name, Control.ClassName])
        );

  for i := 1 to GUI.ComponentCount do
  begin
     if GetFocused = Control then
       EXIT;
     Tab;
  end;
  Fail(Format('%s: Could not Tab to control "%s"', [Msg, Control.Name]), CallerAddr);
end;


procedure TGUITestCase.CheckFocused(Control: TControl);
var
  F :TControl;
begin
  Assert(Control <> nil, 'No control');

  Check(Control is TWinControl,
        Format('Expected a TWinControl, but %s is a %s',
               [Control.Name, Control.ClassName])
        );
  Check(TWinControl(Control).CanFocus,
        Format('Control %s cannot focus', [Control.ClassName])
        );
  if (Control.Owner <> nil) and (Control.Owner is TCustomForm) then
    F := TCustomForm(Control.Owner).ActiveControl
  else
    F := GetFocused;
  if  F <> Control then
  begin
    if F <> nil then
      Fail(Format('Expected control %s to have focus, but %s had it.', [Control.Name, F.Name]), CallerAddr)
    else
      Fail(Format('Expected control %s to have focus', [Control.Name]), CallerAddr);
  end
end;

procedure TGUITestCase.CheckFocused(ControlName: string);
begin
  CheckFocused(FindControl(ControlName));
end;

procedure TGUITestCase.CheckTabTo(ControlName, Msg: string);
begin
  CheckTabTo(FindControl(ControlName), Msg);
end;

end.
